library swg.config.ui.angular.component.end_user_messages;

import 'package:angular/angular.dart';
import 'package:swg_config_ui_angular/src/common/locale.dart';
import 'package:swg_ui_common/src/model/change_action.dart';
import 'package:swg_ui_common/src/model/eum.dart';
import 'package:swg_ui_common/src/service/building_block_service.dart';
import 'package:swg_ui_common/src/util/customer_model_locator.dart';
import 'package:swg_ui_common/src/util/url_validation.dart';
import 'package:tw_portal/portal.dart';
import 'package:tw_ui_angular/tw_ui_angular.dart';

@Component(selector: 'twa-swg-end-user-messages', templateUrl: 'end_user_messages.html', styleUrls: const [
  'end_user_messages.css'
], pipes: const [
  InterpretPipe
], exports: const [
  SwgConfigLocaleInfo,
  ChangeAction
], directives: const [
  CORE_DIRECTIVES,
  twaDirectives
], providers: const [
  const Provider(ComponentService, useFactory: getComponentService),
  const Provider(ComponentBase, useExisting: EndUserMessagesComponent)
])
class EndUserMessagesComponent extends PortalComponent {
  ///
  /// Services & Constructor
  ///

  final Logger _log = new Logger("EndUserMessagesComponent");

  @override
  final List<String> localeRoots = [
    SwgConfigLocaleInfo.PORTAL_KEY,
    SwgConfigLocaleInfo.SWG_CONFIG_KEY,
    SwgConfigLocaleInfo.SWG_CONFIG_COMMON_KEY
  ];

  final BuildingBlockService _buildingBlockService;

  EndUserMessagesComponent(ComponentService cmpSvc, this._buildingBlockService) : super(cmpSvc);

  ///
  /// Template Members
  ///

  final int nameLength = 100;

  List<EndUserMessage> endUserMessages;
  EndUserMessage selectedEndUserMessage;
  bool rightPanelOpened = false;
  bool waiting = false;
  ChangeAction operation;
  int selectedIndex;
  String formHeaderValue = "";
  bool showInUseAlert = false;

  ///
  /// Template Members: DOM element
  ///

  @ViewChild('endUserMessagesGrid')
  DatagridComponent endUserMessagesGrid;

  ///
  /// Template Members: Event Handling
  ///

  onAdd(DatagridAddEvent addEvent) {
    _log.debug("onAdd addEvent $addEvent");
    _resetFormValues();
    formHeaderValue = interpret(SwgConfigLocaleInfo.ADD_END_USER_MESSAGE);
    selectedEndUserMessage = new EndUserMessage();
    operation = ChangeAction.ADD;
    rightPanelOpened = true;
  }

  onDelete(DatagridDeleteEvent deleteEvent) {
    _log.debug("onDelete deleteEvent $deleteEvent");
    var deleteItem = deleteEvent.selectedItem;

    if (deleteItem.inUse) {
      showInUseAlert = true;
      return;
    }

    endUserMessages.remove(deleteItem);
    _resetFormValues();
    _resetDetailViewValues();
  }

  onRefresh(DatagridRefreshEvent refreshEvent) {
    _log.debug("onRefresh refreshEvent $refreshEvent");
    _resetFormValues();
    _resetDetailViewValues();
    _getEndUserMessagesForCustomer();
  }

  onSelectItem(EndUserMessage selectedItem) {
    _log.debug("selectedEndUserMessageChange selected $selectedItem");
    _resetFormValues();
    if (selectedItem != null) {
      _oldEndUserMessage = selectedItem;
      selectedEndUserMessage = _oldEndUserMessage.clone();

      formHeaderValue = interpret(SwgConfigLocaleInfo.EDIT_END_USER_MESSAGE);
      operation = ChangeAction.EDIT;
      rightPanelOpened = true;
      operation = ChangeAction.EDIT;
    }
  }

  onSaveEndUserMessage() {
    if (_oldEndUserMessage == selectedEndUserMessage) {
      _log.debug("+++++ NOTHING CHANGED +++++");
      _resetFormValues();
      _resetDetailViewValues();
      return;
    }

    List copy = new List.from(endUserMessages);
    if (operation == ChangeAction.ADD) {
      copy.add(selectedEndUserMessage);
      copy.sort();
    } else if (operation == ChangeAction.EDIT) {
      copy.remove(_oldEndUserMessage);
      copy.add(selectedEndUserMessage);
    }

    endUserMessages.clear();
    endUserMessages = copy;
    endUserMessages.sort();

    _resetFormValues();
    _resetDetailViewValues();
  }

  onFormReset() {
    _log.debug("onFormReset");
    _resetFormValues();
    _resetDetailViewValues();
  }

  ///
  /// Template Members: Functions
  ///

  String validateEumName(String value) {
    if ((_oldEndUserMessage == null || _oldEndUserMessage.name != selectedEndUserMessage.name) &&
        _checkEumExists(selectedEndUserMessage.name)) {
      return interpret(SwgConfigLocaleInfo.END_USER_MESSAGE_EXISTS);
    }

    return "";
  }

  String validateUrlWithProtocol(String value) {
    if (value == null || value.isEmpty) return "";
    return getUrlWithProtocolValidationMessage(value);
  }

  ///
  /// Lifecycle Hooks
  ///
  @override
  onNgInit() async {
    _log.debug("+++++ onNgInit +++++");
    _loadData();
  }

  ///
  /// Private Members
  ///

  EndUserMessage _oldEndUserMessage;

  _resetFormValues() {
    selectedEndUserMessage = null;
  }

  _resetDetailViewValues() {
    rightPanelOpened = false;
    endUserMessagesGrid.selectedItem = null;
    endUserMessagesGrid.selectedIndex = null;
  }

  _loadData() async {
    _log.debug("_loadData");
    await _getEndUserMessagesForCustomer();
  }

  _getEndUserMessagesForCustomer() async {
    if (CustomerModelLocator.customer != null) {
      List copyList = await _buildingBlockService.getEndUserMessagesForCustomer(CustomerModelLocator.customer.id);
      copyList.sort();
      endUserMessages = copyList ?? [];
    }
  }

  bool _checkEumExists(String name) {
    bool match = false;
    for (EndUserMessage eum in endUserMessages) {
      if (eum.name.toLowerCase() == name.toLowerCase()) {
        match = true;
        break;
      }
    }
    return match;
  }

  // TODO: handle error
  _handleError(e) {
    String msg = (e != null && e.message != null) ? e.message : "";
    _log.error(Error.safeToString(msg), e);
    notifyError(msg);
  }
}
