library swg.config.ui.angular.component.url_lists;

import 'dart:html';

import 'package:angular/angular.dart';
import 'package:swg_config_ui_angular/src/common/locale.dart';
import 'package:swg_ui_common/src/model/change_action.dart';
import 'package:swg_ui_common/src/model/url_item.dart';
import 'package:swg_ui_common/src/model/url_list.dart';
import 'package:swg_ui_common/src/model/url_list_summary.dart';
import 'package:swg_ui_common/src/service/building_block_service.dart';
import 'package:swg_ui_common/src/util/customer_model_locator.dart';
import 'package:swg_ui_common/src/util/punycode.dart';
import 'package:swg_ui_common/src/util/url_list_util.dart';
import 'package:swg_ui_common/src/util/url_validation.dart';
import 'package:tw_portal/portal.dart';
import 'package:tw_ui_angular/tw_ui_angular.dart';

@Component(selector: 'twa-swg-url-lists', templateUrl: 'url_lists.html', styleUrls: const [
  'url_lists.css'
], pipes: const [
  InterpretPipe
], exports: const [
  SwgConfigLocaleInfo,
  ChangeAction
], directives: const [
  CORE_DIRECTIVES,
  twaDirectives
], providers: const [
  const Provider(ComponentService, useFactory: getComponentService),
  const Provider(ComponentBase, useExisting: UrlListsComponent)
])
class UrlListsComponent extends PortalComponent {
  ///
  /// Services & Constructor
  ///

  final Logger _log = new Logger("UrlListsComponent");

  @override
  final List<String> localeRoots = [
    SwgConfigLocaleInfo.PORTAL_KEY,
    SwgConfigLocaleInfo.SWG_CONFIG_KEY,
    SwgConfigLocaleInfo.SWG_CONFIG_COMMON_KEY
  ];

  final BuildingBlockService _buildingBlockService;

  UrlListsComponent(ComponentService cmpSvc, this._buildingBlockService) : super(cmpSvc);

  ///
  /// Template Members
  ///

  final int nameLength = 100;

  List<UrlListSummary> urlListSummaryLists;
  UrlListSummary selectedUrlListSummary;
  bool rightPanelOpened = false;
  bool waiting = false;
  ChangeAction operation;
  int selectedIndex;
  String formHeaderValue = "";

  UrlList selectedUrlList;
  List<UrlItem> urlGridProvider;
  bool urlAddEditDialogVisible = false;
  String dialogLabel;
  ChangeAction urlAction;
  UrlItem selectedUrlItem;
  bool showInUseAlert = false;
  bool urlImportDialogVisible = false;
  bool urlImportErrorVisible = false;

  // urlAddEditDialog
  bool urlItemFormValid = true;

  String get urlItemFormPrimaryState => urlItemFormValid ? ButtonComponent.NORMAL_STATE : ButtonComponent.ERROR_STATE;
  String urlItemFormValidationMessage = "";

  // import dialog
  String selectedRadioOption = "replace";
  List<String> urlImportErrors = [];

  ///
  /// Template Members: DOM element
  ///

  @ViewChild('urlListGrid')
  DatagridComponent urlListGrid;

  @ViewChild('urlAddEditDialog')
  DialogComponent urlAddEditDialog;

  @ViewChild('urlListForm')
  FormComponent urlListForm;

  @ViewChild('urlAddEditForm')
  FormComponent urlAddEditForm;

  @ViewChild('urlGrid')
  DatagridComponent urlGrid;

  @ViewChild('fileInput')
  var fileInput;

  ///
  /// Template Members: Event Handling
  ///

  onAdd(DatagridAddEvent addEvent) {
    _log.debug("onAdd addEvent $addEvent");
    _resetFormValues();
    formHeaderValue = interpret(SwgConfigLocaleInfo.ADD_URL_LIST);
    selectedUrlList = new UrlList();
    operation = ChangeAction.ADD;
    rightPanelOpened = true;
  }

  onDelete(DatagridDeleteEvent deleteEvent) {
    _log.debug("onDelete deleteEvent $deleteEvent");

    if (selectedUrlList.inUse) {
      showInUseAlert = true;
      return;
    }

    _resetFormValues();
    _resetDetailViewValues();
  }

  onRefresh(DatagridRefreshEvent refreshEvent) {
    _log.debug("onRefresh refreshEvent $refreshEvent");
    _resetFormValues();
    _resetDetailViewValues();
    _localUrls.clear();
    _getUrlListSummaryForCustomer();
  }

  onSelectItem(UrlListSummary selectedItem) {
    _log.debug("selectedUrlListSummaryChange selected $selectedItem");
    _resetFormValues();
    if (selectedItem != null) {
      selectedUrlListSummary = selectedItem;
      _populateUrlList();

      formHeaderValue = interpret(SwgConfigLocaleInfo.EDIT_URL_LIST);
      operation = ChangeAction.EDIT;
      rightPanelOpened = true;
    }
  }

  onSelectItemIndex(int selectedItemIndex) {
    selectedIndex = selectedItemIndex;
  }

  onSubmitUrlListForm() {
    if (selectedUrlList == _oldUrlList) {
      _log.debug("+++++ NOTHING CHANGED +++++");
      _resetFormValues();
      _resetDetailViewValues();
      return;
    }

    List copy = new List.from(urlListSummaryLists);
    if (operation == ChangeAction.ADD) {
      copy.add(new UrlListSummary.fromUrlListVO(selectedUrlList));
      copy.sort();
    } else if (operation == ChangeAction.EDIT) {
      copy.remove(selectedUrlListSummary);
      copy.add(new UrlListSummary.fromUrlListVO(selectedUrlList));
    }

    urlListSummaryLists.clear();
    urlListSummaryLists = copy;
    urlListSummaryLists.sort();
    _localUrls[selectedUrlList.id] = urlGridProvider != null ? new List.from(urlGridProvider) : [];

    _resetFormValues();
    _resetDetailViewValues();
  }

  onResetUrlListForm() {
    _log.debug("onFormReset");
    _resetFormValues();
    _resetDetailViewValues();
  }

  onAddUrl(DatagridAddEvent addEvent) {
    dialogLabel = interpret(SwgConfigLocaleInfo.ADD_URL);
    selectedUrlItem = new UrlItem();
    urlAction = ChangeAction.ADD;
    urlAddEditDialogVisible = true;
  }

  onEditUrl(DatagridEditEvent editEvent) {
    _oldUrlItem = editEvent.selectedItem;
    selectedUrlItem = _oldUrlItem.clone();
    if (selectedUrlItem.url.contains("xn--")) {
      selectedUrlItem.url = idnaUrlDecoder(selectedUrlItem.url);
    }
    dialogLabel = interpret(SwgConfigLocaleInfo.EDIT_URL);
    urlAction = ChangeAction.EDIT;
    urlAddEditDialogVisible = true;
  }

  onDeleteUrl(DatagridDeleteEvent deleteEvent) {
    if (urlGrid.checkedItems != null && urlGrid.checkedItems.isNotEmpty) {
      List<UrlItem> copyList = new List.from(urlGridProvider);
      urlGrid.checkedItems.forEach((toDelete) {
        copyList.removeWhere((url) => url.id == (toDelete as UrlItem).id);
      });
      urlGridProvider = copyList;
      selectedUrlList.urls = urlGridProvider;
    }
  }

  onSubmitUrlAddEdit() {
    if (!urlAddEditForm.validate()) {
      urlItemFormValid = false;
      urlItemFormValidationMessage = interpret(SwgConfigLocaleInfo.INVALID_FORM);
      return;
    }
    if (urlAction == ChangeAction.ADD) {
      selectedUrlItem.url = idnaUrlEncoder(selectedUrlItem.url);
      List copy = new List.from(urlGridProvider);
      copy.add(selectedUrlItem.clone());
      copy.sort((a, b) => a.url.toLowerCase().compareTo(b.url.toLowerCase()));
      urlGridProvider.clear();
      urlGridProvider = copy;
    } else if (urlAction == ChangeAction.EDIT) {
      selectedUrlItem.url = idnaUrlEncoder(selectedUrlItem.url);
      List copy = new List.from(urlGridProvider);
      copy.removeWhere((url) => url.id == selectedUrlItem.id);
      copy.add(selectedUrlItem.clone());
      copy.sort((a, b) => a.url.toLowerCase().compareTo(b.url.toLowerCase()));
      urlGridProvider.clear();
      urlGridProvider = copy;
    }
    selectedUrlList.urls = urlGridProvider;
    onCancelUrlAddEdit();
  }

  onCancelUrlAddEdit() {
    urlAddEditForm.reset();
    urlItemFormValid = true;
    urlItemFormValidationMessage = "";
    urlAddEditDialogVisible = false;
  }

  onImportListButtonClick() {
    _log.debug("onImportListButtonClick");
    urlImportDialogVisible = true;
  }

  onSubmitUrlListImport() {
    String importMode = selectedRadioOption;
    urlImportErrors = [];

    FileList files = fileInput.nativeElement.files;
    File file = files.item(0);

    if (file != null) {
      if (file.name != null && (file.name.contains("text") || file.name.contains("txt"))) {
        FileReader reader = new FileReader();
        reader.onLoad.listen((e) {
          String fileContents = reader.result.toString();
          var importList;
          try {
            importList = _readUrlFileContents(fileContents);
          } catch (FormatException) {
            urlImportErrorVisible = true;
            return;
          }
          _completeImport(importList, importMode);
          urlImportDialogVisible = false;
        });
        reader.readAsText(file);
      } else {
        String fileType = "";
        List fileTokens = file.name.split(".");
        if (fileTokens.length > 1) {
          fileType = fileTokens[1];
        }
        urlImportErrors.add(interpret(SwgConfigLocaleInfo.INVALID_FILE_TYPE) + ": $fileType");
        urlImportErrorVisible = true;
      }
    } else {
      urlImportErrors.add(interpret(SwgConfigLocaleInfo.NO_FILE_SELECTED));
      urlImportErrorVisible = true;
    }
  }

  onResetUrlImport() {
    urlImportDialogVisible = false;
  }

  onSubmitUrlImportError() {
    urlImportErrorVisible = false;
  }

  ///
  /// Template Members: Functions
  ///

  String validateUrlListName(String value) {
    if (urlListSummaryLists.any((summary) =>
        summary.name.toLowerCase() == value.toLowerCase() && selectedUrlList.id != (summary as UrlListSummary).id)) {
      return interpret(SwgConfigLocaleInfo.URL_LIST_NAME_EXISTS);
    }
    return "";
  }

  String validateUrlName(String value) {
    String validationMessage = getUrlValidationMessage(value);
    if (validationMessage.isNotEmpty) {
      return validationMessage;
    }
    if (urlExistsInList(value, urlGridProvider) && !_editingUrlWithName(value)) {
      return interpret(SwgConfigLocaleInfo.URL_NAME_EXISTS);
    }
    return "";
  }

  String validateUrlWithProtocol(String value) {
    if (value == null || value.isEmpty) return "";
    return getUrlWithProtocolValidationMessage(value);
  }

  bool readOnlyName() {
    return selectedUrlList != null && UrlListUtil.isSystemGenerated(selectedUrlList.name);
  }

  ///
  /// Lifecycle Hooks
  ///
  @override
  onNgInit() async {
    _log.debug("+++++ onNgInit +++++");
    _loadData();
  }

  ///
  /// Private Members
  ///

  UrlList _oldUrlList;
  UrlItem _oldUrlItem;
  Map<String, List<UrlItem>> _localUrls = {};

  _resetFormValues() {
    selectedUrlListSummary = null;
    urlGridProvider = null;
    selectedUrlList = null;
    selectedUrlItem = null;
    _oldUrlList = null;
    _oldUrlItem = null;
  }

  _resetDetailViewValues() {
    rightPanelOpened = false;
    urlListGrid.selectedItem = null;
    urlListGrid.selectedIndex = null;
  }

  _loadData() async {
    _log.debug("_loadData");
    await _getUrlListSummaryForCustomer();
  }

  _getUrlListSummaryForCustomer() async {
    if (CustomerModelLocator.customer != null) {
      List copyList = await _buildingBlockService.getUrlListSummariesForCustomer(CustomerModelLocator.customer.id);
      copyList.sort();
      urlListSummaryLists = copyList ?? [];
    }
  }

  bool _editingUrlWithName(String url) {
    if (urlAction != ChangeAction.EDIT) {
      return false;
    }
    var urlBeingEdited = urlGridProvider.singleWhere((urlItem) => urlItem.id == selectedUrlItem.id);
    return urlBeingEdited.normalized == normalizeUrl(url);
  }

  _populateUrlList() async {
    if (selectedUrlListSummary != null && selectedUrlListSummary.name != null) {
      UrlListSummary summary = selectedUrlListSummary;
      if (_localUrls.containsKey(summary.id)) {
        _oldUrlList = new UrlList(summary.id, summary.name, summary.notes, _localUrls[summary.id]);
        selectedUrlList = _oldUrlList.clone();
      } else {
        try {
          _oldUrlList = await _buildingBlockService.getUrlList(CustomerModelLocator.customer.id, summary.id);
          selectedUrlList = _oldUrlList.clone();
          _localUrls[selectedUrlList.id] = selectedUrlList.urls;
        } catch (error) {
          _handleError(error);
        }
      }
      urlGridProvider = selectedUrlList.urls;
    }
  }

  List<String> _readUrlFileContents(String fileContents) {
    var importList = [];
    var lines = fileContents.split("\n");
    for (int i = 0; i < lines.length; i++) {
      var line = lines[i];
      if (line.isEmpty) {
        continue;
      }
      var urlComponents = line.split(",");
      String url = urlComponents[0].trim();
      String notes;
      String validationMessage = getUrlValidationMessage(url);
      if (validationMessage.isNotEmpty) {
        urlImportErrors.add("Line ${i + 1}: $validationMessage - $url");
        continue;
      }
      if (urlExistsInList(url, importList)) {
        urlImportErrors.add("Line ${i + 1}: duplicate URL - $url");
      }
      if (urlComponents.length > 1) {
        // this join bit accounts for notes containing commas
        notes = urlComponents.sublist(1).join(',').trim();
      }
      UrlItem itemVO = new UrlItem()
        ..url = url
        ..notes = notes;
      importList.add(itemVO);
    }
    if (urlImportErrors.isNotEmpty) {
      throw new FormatException("Couldn't parse URL file");
    } else {
      return importList;
    }
  }

  _completeImport(List importList, String importMode) {
    List copy;
    if (importMode == "merge") {
      copy = new List.from(urlGridProvider);
      for (UrlItem urlItem in importList) {
        if (urlExistsInList(urlItem.url, urlGridProvider)) {
          UrlItem item = urlGridProvider.singleWhere((e) => e.normalized == urlItem.normalized);
          urlItem.id = item.id;
          copy.removeWhere((url) => url.id == urlItem.id);
          copy.add(urlItem.clone());
        } else {
          copy.add(urlItem.clone());
        }
      }
    } else if (importMode == "replace") {
      copy = new List.from(importList);
    }
    copy.sort((a, b) => a.url.toLowerCase().compareTo(b.url.toLowerCase()));
    urlGridProvider = copy;
  }

  // TODO: handle error
  _handleError(e) {
    String msg = (e != null && e.message != null) ? e.message : "";
    _log.error(Error.safeToString(msg), e);
    notifyError(msg);
  }
}
