library swg.config.ui.angular.component.locations;

import 'dart:html';
import 'package:angular/angular.dart';
import 'package:angular_components/angular_components.dart';
import 'package:swg_config_ui_angular/src/common/locale.dart';
import 'package:swg_config_ui_angular/src/ui/component/identities/base_identities.dart';
import 'package:swg_config_ui_angular/src/ui/component/rule/rule_common.dart';
import 'package:swg_ui_common/src/model/change_action.dart';
import 'package:swg_ui_common/src/model/customer_location.dart';
import 'package:swg_ui_common/src/model/identifier_type.dart';
import 'package:swg_ui_common/src/model/identity.dart';
import 'package:swg_ui_common/src/model/location.dart';
import 'package:swg_ui_common/src/service/identifier_cache.dart';
import 'package:swg_ui_common/src/service/identity_service.dart';
import 'package:swg_ui_common/src/util/customer_model_locator.dart';
import 'package:swg_ui_common/src/util/ip_validation.dart';
import 'package:tw_portal/portal.dart';
import 'package:tw_ui_angular/tw_ui_angular.dart';

@Component(selector: 'twa-swg-locations',
    templateUrl: 'locations.html',
    styleUrls: const ['locations.css'],
    pipes: const [InterpretPipe],
    exports: const [SwgConfigLocaleInfo, ChangeAction],
    directives: const [twaDirectives, RulesComponent],
    providers: const [const Provider(ComponentService, useFactory: getComponentService),
    const Provider(ComponentBase, useExisting: LocationsComponent),
    swgIdentityServiceProviders, identifierCacheProviders]
)
class LocationsComponent extends PortalComponent with BaseIdentitiesComponent {

  final Logger _log = new Logger("LocationsComponent");

  static const String locationErrorMessage = 'Error loading Locations';

  LocationsComponent(ComponentService cmpSvc, this._identityService, this._identifierCache) : super(cmpSvc);

  final RpcIdentityService _identityService;
  final IdentifierCache _identifierCache;

  @override
  final List<String> localeRoots = [ SwgConfigLocaleInfo.PORTAL_KEY, SwgConfigLocaleInfo.SWG_CONFIG_KEY, SwgConfigLocaleInfo.SWG_CONFIG_COMMON_KEY];

  String redirectOption;
  String from;
  String to;
  String predefinedSite;
  String get redirectSiteSelectLabel => redirectSiteSelectionModel.selectedValues.length > 0
      ? redirectSiteSelectionModel.selectedValues.first
      : "";
  List predefinedLocalSites;
  bool enableRedirect = false;
  bool siteRequired = false;
  RegExp badChars = new RegExp(r'[,±~#$%^&(){}[\]"?><*+=|;]');

  SelectionModel redirectSiteSelectionModel = new SelectionModel.withList(allowMulti: false, selectedValues: []);
  SelectionOptions<String> get predefinedLocalSitesOptions => new SelectionOptions.fromList(predefinedLocalSites ?? []);

  @override
  onNgInit() async {
    _loadData();
  }

  onAdd(DatagridAddEvent addEvent) {
    _log.debug("onAdd addEvent $addEvent");
    formHeaderValue = interpret(SwgConfigLocaleInfo.ADD_LOCATION);
    _resetFormValues();
    newIdentity = new LocationIdentity();
    operation = ChangeAction.ADD;
    isEditing = false;
    rightPanelOpened = true;
    nothingSelected = false;
  }

  onDelete(DatagridDeleteEvent deleteEvent) {
    _log.debug("onDelete DatagridDeleteEvent $deleteEvent");
    identities.remove(deleteEvent.selectedItem);
    _resetFormValues();
    _resetDetailViewValues();
  }

  onRefresh(DatagridRefreshEvent refreshEvent) {
    _resetFormValues();
    rightPanelOpened = false;
    nothingSelected = true;
    _loadData();
  }

  onSelectItem(LocationIdentity selectedItem) {
    _log.debug("onSelectItem selected $selected");
    _resetFormValues();
    if (selectedItem != null) {
      selected = selectedItem;
      newIdentity = selected.clone();
      formHeaderValue = interpret(SwgConfigLocaleInfo.EDIT_LOCATION);
      identifier = newIdentity.identifier;
      enableRedirect = (newIdentity as LocationIdentity).enableRedirect;
      redirectOption = (newIdentity as LocationIdentity).redirectOption;
      (newIdentity as LocationIdentity).redirectLocalSite != null ?
      redirectSiteSelectionModel.select((newIdentity as LocationIdentity).redirectLocalSite) : null;
      from = (newIdentity as LocationIdentity).from;
      to = (newIdentity as LocationIdentity).to;
      newIdentity.policyIdentifier != null ? policySelectionModel.select(newIdentity.policyIdentifier) : '';
      newIdentity.timezone != null ? timezoneSelectionModel.select(newIdentity.timezone) : '';
      newIdentity.safeSearch != null ? safeSearchSelectionModel.select(newIdentity.safeSearch) : '';
      identityRules = new List.from(selected.identityRules ?? []);
      operation = ChangeAction.EDIT;
      isEditing = true;
      rightPanelOpened = true;
      nothingSelected = false;
      operation = ChangeAction.EDIT;
    }
  }

  String validateRangeName(String value) {
    String msg = "";
    if (!isEditing && (identifier == null || identifier.isEmpty)) {
      //set('newIdentity.identifier', newIdentity.from);
      identifier = from;
    }
    if (!isEditing && identities.any((LocationIdentity vo) => vo.identifier == value)) {
      return interpret(SwgConfigLocaleInfo.LOCATION_EXISTS);
    }
    if (identifier.contains(" ")) {
      return interpret(SwgConfigLocaleInfo.RANGE_NAME_NO_SPACES);
    }
    return validateIdentityName(identifier);
  }

  onSelectItemIndex(int selectedItemIndex) {
    selectedIndex = selectedItemIndex;
  }

  redirectOptionRadioChanged(RadioGroupEvent e) {
    _log.debug("redirectOptionRadioChanged ${e.radioGroup.value}");
    if (e != null && e.radioGroup != null && e.radioGroup.value != null) {
      e.radioGroup.value ? siteRequired = true : siteRequired = false;
    }
  }

  bool isFromAddressDisabled(bool editing, String identity, String fromAddress, String toAddress) {
    return editing && (fromAddress == identifier) && (toAddress == null || toAddress.isEmpty);
  }

  bool isRangeNameRequired(bool editing, String toAddress) {
    return !editing && toAddress != null && toAddress.isNotEmpty;
  }

  /*Future<String> validateLocationForm() async {
    String msg = validate_ipv6(from);
    if (msg.isEmpty) {
      from = collapseIPv6(from);
      if (to != null && to.isNotEmpty) {
        msg = validate_ipv6(to);
        if (msg.isEmpty) {
          to = collapseIPv6(to);
          msg = await _validateIpForm();
        }
      }
    }
    return msg;
  }*/

  Future<String> validateLocationForm() async {
    if (validate_ipv6(from).isEmpty) {
      from = collapseIPv6(from);
    }
    if (to != null && to.isNotEmpty && validate_ipv6(to).isEmpty) {
      to = collapseIPv6(to);
    }

    String msg = await _validateIpForm();
    return msg;
  }

  onFormSubmit() async {
    _log.debug("onFormSubmit");
    if (!isEditing && (identifier == null || identifier.isEmpty)) {
      identifier = from;
    }

    formMessage = await validateLocationForm();
    if (formMessage.isNotEmpty) {
      formState = ButtonComponent.ERROR_STATE;
      locationsForm.setFormState(error : true, message : formMessage);
      return;
    }

    newIdentity.identifier = identifier;
    newIdentity.policyIdentifier = getDropDownSelectedItem(policySelectionModel);
    newIdentity.timezone = getDropDownSelectedItem(timezoneSelectionModel);
    newIdentity.safeSearch = getDropDownSelectedItem(safeSearchSelectionModel);
    newIdentity.identityRules = new List.from(identityRules ?? []);
    (newIdentity as LocationIdentity).from = from;
    (newIdentity as LocationIdentity).to = to;
    (newIdentity as LocationIdentity).enableRedirect = enableRedirect;
    // These two are causing the serilization issue , check again
    (newIdentity as LocationIdentity).redirectOption = redirectOption;
    (newIdentity as LocationIdentity).redirectLocalSite = getDropDownSelectedItem(redirectSiteSelectionModel);

    Identity identity = await _identityService.getIdentityWithSameName(CustomerModelLocator.customer.id, newIdentity);
    //bool isIdentityNameInUse = _isIdentityWithSameNameExists(identity);
    if (identity != null) {
      formMessage = interpret(SwgConfigLocaleInfo.IDENTITY_WITH_SAME_NAME_AND_DIFFERENT_TYPE_EXISTS);
      formState = ButtonComponent.ERROR_STATE;
      locationsForm.setFormState(error : true, message : formMessage);
      return;
    }


    List identitiesCopy = new List.from(identities);
    if (operation == ChangeAction.ADD) {
      identitiesCopy.add(newIdentity);
      identities.clear();
      identities = identitiesCopy;
    } else {
      identitiesCopy.remove(selected);
      identitiesCopy.add(newIdentity);
      identities.clear();
      identities = identitiesCopy;
    }

    ChangeAction action = isEditing ? ChangeAction.EDIT : ChangeAction.ADD;
    if (selected != null && selected.policyIdentifier != newIdentity.policyIdentifier) {
      newIdentity.parentUpdated = true;
    }

    //handleMscSettings();

    if ((selected == null && newIdentity.policyIdentifier != null)
        || (selected != null && selected.policyIdentifier != newIdentity.policyIdentifier)) {
      identityUpdatedOrDeleted = true;
    }
    _log.debug("selectedIndex $selectedIndex newIdentity $newIdentity");

    // this is not working
    //userIdentitiesGrid.updateRow(selectedIndex, newIdentity);
    _resetFormValues();
    _resetDetailViewValues();
  }

  @ViewChild('locationsForm')
  FormComponent locationsForm;

  Future<String> _validateIpForm() async {
    String msg = "";
    to ??= "";
    if (to.isNotEmpty) {
      if (!_isValidIpRange(from, to)) {
        return interpret(SwgConfigLocaleInfo.FROM_AND_TO_SAME_TYPE);
      }
      if (!_isRangeInOrder(from, to)) {
        return interpret(SwgConfigLocaleInfo.RANGE_START_LOWER_RANGE_END);
      }
    }
    // only need to check for overlapping locations if adding new location or 'from' or 'to' has been edited
    if (!isEditing || from != (selected as LocationIdentity).from
        || (to ?? "") != ((selected as LocationIdentity).to ?? "")) {
      String overlappingErrorMessage = await _getOverlappingErrorMessage(from, to);
      if (overlappingErrorMessage.isNotEmpty) {
        msg = overlappingErrorMessage;
        return msg;
      }
    }
    return msg;
  }

  bool _isValidIpRange(String from, String to) {
    return _valid_ipv4_range(from, to) || _valid_ipv6_range(from, to);
  }

  bool _isRangeInOrder(String from, String to) {
    List fromByteList;
    List toByteList;
    if (_valid_ipv4_range(from, to)) {
      // assume ipv4 bytes are . separated, and in decimal format
      fromByteList = from.split('.').map((x) => (int.parse(x))).toList();
      toByteList = to.split('.').map((x) => (int.parse(x))).toList();
    } else if (_valid_ipv6_range(from, to)) {
      // assume ipv6 addresses are in hex, : separated, and potentially collapsed
      fromByteList = from.split(':').map((x) => (int.parse(x.isEmpty ? '0' : x, radix: 16))).toList();
      toByteList = to.split(':').map((x) => (int.parse(x.isEmpty ? '0' : x, radix: 16))).toList();
    } else {
      // not a valid ip range
      return false;
    }
    for(int i = 0; i < fromByteList.length; i++) {
      if (fromByteList [i] < toByteList[i]) {
        return true;
      } else if (fromByteList [i] > toByteList[i]) {
        return false;
      }
    }
    // this is only reached if the ips are equal
    return false;
  }

  bool _valid_ipv4_range(String from, String to) {
    return validate_ipv4(from).isEmpty && validate_ipv4(to ?? "").isEmpty;
  }

  bool _valid_ipv6_range(String from, String to) {
    return validate_ipv6(from).isEmpty && validate_ipv6(to ?? "").isEmpty;
  }

  Future<String> _getOverlappingErrorMessage(String from, String to) async {
    try {
      List<CustomerLocation> overlappingLocations = await identityService
          .getOverlappingCustomerLocations(from, to);

      // remove the location currently being edited - can never be in conflict with itself
      overlappingLocations?.removeWhere((location) => location.customerId == CustomerModelLocator.customer.id
          && location.locationIdentifier == newIdentity.identifier);

      if (overlappingLocations == null || overlappingLocations.isEmpty) {
        return "";
      }

      // any overlapping location that belongs to another customer is a conflict
      List<CustomerLocation> conflictingLocations = new List.from(overlappingLocations.where((location) => location.customerId != CustomerModelLocator.customer.id));
      // remove all overlapping locations that don't belong to this customer
      overlappingLocations.removeWhere((location) => location.customerId != CustomerModelLocator.customer.id);

      if (to == null || to.isEmpty) {
        // single IP address cannot overlap with other single IPs (can overlap with ranges)
        List conflictingIPs = overlappingLocations.where((location) => location.ipTo == null || location.ipTo.isEmpty).toList();
        if (conflictingIPs != null && conflictingIPs.isNotEmpty) {
          conflictingIPs.forEach((ip) => conflictingLocations.add(ip));
        }
      }
      else {
        // ranges cannot overlap with other ranges (can overlap with single IPs)
        List conflictingRanges = overlappingLocations.where((location) => location.ipTo != null && location.ipTo.isNotEmpty).toList();
        if (conflictingRanges != null && conflictingRanges.isNotEmpty) {
          conflictingRanges.forEach((ip) => conflictingLocations.add(ip));
        }
      }

      if (conflictingLocations != null && conflictingLocations.isNotEmpty) {
        StringBuffer sb = new StringBuffer();
        sb.write(interpret(SwgConfigLocaleInfo.IP_OVERLAPPING_PART1)+"${conflictingLocations.length > 1 ?
        interpret(SwgConfigLocaleInfo.PLURAL_CHARACTER) : ""}"+ ":");
        for (CustomerLocation location in conflictingLocations) {
          sb.write("\n"+interpret(SwgConfigLocaleInfo.CUSTOMER) + ":" +"${location.customerName},"
              + interpret(SwgConfigLocaleInfo.LOCATION) + ":" + "${location.ipFrom}"+
              "${location.ipTo != null && location.ipTo.isNotEmpty ? "-${location.ipTo}" : ""}");
        }
        return sb.toString();
      }
      return "";
    } catch (_) {
      notifyError(locationErrorMessage);
      return "Error validating location";
    }
  }

  onFormReset() {
    _log.debug("onFormReset");
  }

  enableRedirectCheckBoxClicked(bool value) {
    enableRedirect = value;
    if (enableRedirect) {
      if (redirectOption == null) {
        redirectOption = "false";
        siteRequired = false;
      }
    } else {
      siteRequired = false;
      redirectSiteSelectionModel = new SelectionModel.withList(allowMulti: false, selectedValues: []);
      redirectOption = null;
    }
  }

  _resetFormValues() {
    identifier = "";
    from = null;
    to = null;
    identityRules = [];
    policySelectionModel = new SelectionModel.withList(allowMulti: false, selectedValues: []);
    safeSearchSelectionModel = new SelectionModel.withList(allowMulti: false, selectedValues: []);
    timezoneSelectionModel = new SelectionModel.withList(allowMulti: false, selectedValues: []);
  }

  _resetDetailViewValues() {
    rightPanelOpened = false;
    nothingSelected = true;
  }

  _loadData() async {
    _log.debug("_loadData");
    await _loadIdentities();
    await _loadPredefinedSites();
    _loadTimeZones();
    _loadPolicies();
    _loadPredefinedLocalSites();
  }

  _loadPolicies() async {
    List policies = new List.from(await _identifierCache.getItemsOfType(IdentifierType.NAMED_POLICY));
    policies.sort();
    namedPolicies = policies;
  }

  _loadIdentities() async {
    try {
      waiting = true;
      List list = await _identityService.getLocationsForCustomer(CustomerModelLocator?.customer.id);
      identities = list ?? [];
      _log.debug("_loadData locations $identities");
    } catch (e) {
      _handleError(e);
      identities = [];
    } finally {
      waiting = false;
    }
  }

  _loadTimeZones() async {
    timeZonesDataProvider = await _identifierCache.getTimeZones();
  }

  _loadPredefinedSites() async {
    List<String> predefinedSites = await _identityService.getPredefinedSites();
    predefinedSites = predefinedSites ?? [];
    if (predefinedSites != null && predefinedSites.isNotEmpty) {
      predefinedSite = predefinedSites.first;
    }
  }

  _loadPredefinedLocalSites() async {
    List localSites = await _identityService.getPredefinedLocalSites(predefinedSite);
    predefinedLocalSites = localSites ?? [];
  }

  _handleError(e) {
    String msg = (e != null && e.message != null) ? e.message : "";
    _log.error(Error.safeToString(msg), e);
    notifyError(msg);
  }

}